unit _Preproc;

/////////////////////////////////////////////////////////////////////
//
// Hi-Track Version 2
// Copyright (c) 2000-2002 Dmitry Liman [2:461/79]
//
// http://hi-files.by.ru
// http://hi-files.narod.ru
//
/////////////////////////////////////////////////////////////////////

interface

uses
  Objects, _MapFile;

const
  STACK_DEPTH = 20;

type
  TFrameMode = (fm_if, fm_else, fm_endif);
  TFrameSet  = set of TFrameMode;
  TOperator  = (op_none, op_ifdef, op_ifndef, op_else, op_endif,
                op_define, op_undef, op_iftime, op_ifflag, op_raise, op_drop);

  PStack = ^TStack;
  TStack = object (TObject)
    constructor Init;
    procedure Push( f: TFrameMode );
    procedure Pop;
    function  Top: TFrameMode;
    function  Level: Integer;
  private
    Items: array [0..STACK_DEPTH] of TFrameMode;
    SP   : Integer;
  end; { TStack }

  PDefines = ^TDefines;
  TDefines = object (TSortedCollection)
    constructor Init;
    procedure FreeItem( Item: Pointer ); virtual;
    function  KeyOf( Item: Pointer ) : Pointer; virtual;
    function  Compare( Key1, Key2: Pointer ) : Integer; virtual;
  end; { TDefines }

  PPreprocessor = ^TPreprocessor;
  TPreprocessor = object (TCollection)
    constructor Init;
    destructor Done; virtual;

    procedure FreeItem( Item: Pointer ); virtual;
    procedure Define( const Key, Val: String );
    procedure Undef ( const Key: String );
    function  IfDef( const Key: String ) : Boolean;

    procedure Exec( const FileName: String );
    function  GetLine( var S: String ) : Boolean;

  private
    Source : TMappedFile;
    Frame  : PStack;
    Defines: PDefines;
    Finger : Integer;

    function What( const S: String ) : TOperator;

    procedure __if( Cond: Boolean );
    procedure __else;
    procedure __endif;

    procedure Enter( f: TFrameMode );
    procedure Leave;
    function  Jump( over: TFrameSet ) : TFrameMode;

    function  Expand( S: String ) : String;
  end; { TPreprocessor }

{ =================================================================== }

implementation

uses SysUtils, MyLib, DT;

////////////////////////////////////////////////////////////////
//
// TStack
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Init                                                      }
{ --------------------------------------------------------- }

constructor TStack.Init;
begin
  inherited Init;
  SP := -1;
end; { Init }

{ --------------------------------------------------------- }
{ Push                                                      }
{ --------------------------------------------------------- }

procedure TStack.Push( f: TFrameMode );
begin
  if SP >= STACK_DEPTH then
    raise Exception.Create( 'Stack overflow' );
  Inc( SP );
  Items[SP] := f;
end; { Push }

{ --------------------------------------------------------- }
{ Pop                                                       }
{ --------------------------------------------------------- }

procedure TStack.Pop;
begin
  if SP < 0 then
    raise Exception.Create( 'Stack underflow' );
  Dec( SP );
end; { Pop }

{ --------------------------------------------------------- }
{ Top                                                       }
{ --------------------------------------------------------- }

function TStack.Top: TFrameMode;
begin
  if SP < 0 then
    raise Exception.Create( 'Empty stack calling' );
  Result := Items[SP];
end; { Top }

{ --------------------------------------------------------- }
{ Level                                                     }
{ --------------------------------------------------------- }

function TStack.Level: Integer;
begin
  Result := SP;
end; { Level }


////////////////////////////////////////////////////////////////
//
// TDefines
//
////////////////////////////////////////////////////////////////

type
  PDefine = ^TDefine;
  TDefine = record
    Key, Val: String;
  end; { TDefine }

{ --------------------------------------------------------- }
{ Init                                                      }
{ --------------------------------------------------------- }

constructor TDefines.Init;
begin
  inherited Init(20, 20);
end; { Init }

{ --------------------------------------------------------- }
{ FreeItem                                                  }
{ --------------------------------------------------------- }

procedure TDefines.FreeItem( Item: Pointer );
begin
  Dispose( PDefine(Item) );
end; { FreeItem }

{ --------------------------------------------------------- }
{ KeyOf                                                     }
{ --------------------------------------------------------- }

function TDefines.KeyOf( Item: Pointer ) : Pointer;
begin
  Result := @PDefine(Item)^.Key;
end; { KeyOf }

{ --------------------------------------------------------- }
{ Compare                                                   }
{ --------------------------------------------------------- }

function TDefines.Compare( Key1, Key2: Pointer ) : Integer;
begin
  Result := cmpi( VString(Key1)^, VString(Key2)^ );
end; { Compare }


////////////////////////////////////////////////////////////////
//
// TPreprocessor
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Init                                                      }
{ --------------------------------------------------------- }

constructor TPreprocessor.Init;
begin
  inherited Init(100, 100);
  Frame := New( PStack, Init );
  Defines := New( PDefines, Init );
end; { Init }

{ --------------------------------------------------------- }
{ Done                                                      }
{ --------------------------------------------------------- }

destructor TPreprocessor.Done;
begin
  Destroy( Defines );
  Destroy( Frame );
  inherited Done;
end; { Done }

{ --------------------------------------------------------- }
{ FreeItem                                                  }
{ --------------------------------------------------------- }

procedure TPreprocessor.FreeItem( Item: Pointer );
begin
  FreeStr( VString(Item) );
end; { FreeItem }


{ --------------------------------------------------------- }
{ Define                                                    }
{ --------------------------------------------------------- }

procedure TPreprocessor.Define( const Key, Val: String );
var
  D: PDefine;
begin
  if Key = '' then
    raise Exception.Create( 'Bad #define' );
  New( D );
  D^.Key := Key;
  D^.Val := Val;
  Defines^.Insert( D );
end; { Define }

{ --------------------------------------------------------- }
{ IfDef                                                     }
{ --------------------------------------------------------- }

function TPreprocessor.IfDef( const Key: String ) : Boolean;
var
  j: Integer;
begin
  Result := Defines^.Search( @Key, j );
end; { IfDef }

{ --------------------------------------------------------- }
{ Undef                                                     }
{ --------------------------------------------------------- }

procedure TPreprocessor.Undef( const Key: String );
var
  j: Integer;
begin
  if Defines^.Search( @Key, j ) then
    Defines^.AtFree( j );
end; { Undef }

{ --------------------------------------------------------- }
{ What                                                      }
{ --------------------------------------------------------- }

function TPreprocessor.What( const S: String ) : TOperator;
begin
  if (S <> '') and (S[1] = '#') then
  begin
    if cmpi(S, '#ifdef') = 0 then
      Result := op_ifdef
    else if cmpi(S, '#ifndef') = 0 then
      Result := op_ifndef
    else if cmpi(S, '#else') = 0 then
      Result := op_else
    else if cmpi(S, '#endif') = 0 then
      Result := op_endif
    else if cmpi(S, '#define') = 0 then
      Result := op_define
    else if cmpi(S, '#undef') = 0 then
      Result := op_undef
    else if cmpi(S, '#iftime') = 0 then
      Result := op_iftime
    else if cmpi(S, '#ifflag') = 0 then
      Result := op_ifflag
    else if cmpi(S, '#drop' ) = 0 then
      Result := op_drop
    else if cmpi(S, '#raise') = 0 then
      Result := op_raise
    else
      raise Exception.Create( 'Bad preprocessor directive: "' + S + '"' );
  end
  else
    Result := op_none;
end; { What }

{ --------------------------------------------------------- }
{ Exec                                                      }
{ --------------------------------------------------------- }

procedure TPreprocessor.Exec( const FileName: String );
var
  S: String;
  Key: String;
  Val: String;
begin
  Finger := 0;
  Source.Init( FileName );
  try
    while Source.GetLine( S ) do
    begin
      SplitPair( S, Key, Val );
      case What( Key ) of
        op_ifdef:
          __if( ifdef(Val) );

        op_ifndef:
          __if( not ifdef(Val) );

        op_else:
          __else;

        op_endif:
          __endif;

        op_define:
          begin
            S := Val;
            SplitPair( S, Key, Val );
            Define( Key, Val );
          end;

        op_undef:
          Undef( Val );

        op_iftime:
          __if( MatchTime(Val) );

        op_ifflag:
          __if( FileExists(Val) );

        op_drop:
          DeleteFile( Val );

        op_raise:
          CreateSemaphore( Val );

      else
        Insert( AllocStr(S) );
      end;

    end;
  finally
    Source.Done;
  end;
end; { Exec }


{ --------------------------------------------------------- }
{ __if                                                      }
{ --------------------------------------------------------- }

procedure TPreprocessor.__if( Cond: Boolean );
begin
  Enter( fm_if );
  if not Cond then
  begin
    if Jump( [fm_else, fm_endif] ) = fm_else then
      Enter( fm_else );
  end;
end; { __if }


{ --------------------------------------------------------- }
{ __else                                                    }
{ --------------------------------------------------------- }

procedure TPreprocessor.__else;
begin
  if Frame^.Top = fm_if then
    Jump( [fm_endif] )
  else
    raise Exception.Create( 'Misplaced #else' );
end; { __else }


{ --------------------------------------------------------- }
{ __endif                                                   }
{ --------------------------------------------------------- }

procedure TPreprocessor.__endif;
begin
  if Frame^.Top in [fm_if, fm_else] then
    Leave
  else
    raise Exception.Create( 'Misplaced #endif' );
end; { __endif }


{ --------------------------------------------------------- }
{ Enter                                                     }
{ --------------------------------------------------------- }

procedure TPreprocessor.Enter( f: TFrameMode );
begin
  Frame^.Push( f );
end; { Enter }


{ --------------------------------------------------------- }
{ Leave                                                     }
{ --------------------------------------------------------- }

procedure TPreprocessor.Leave;
begin
  Frame^.Pop;
end; { Leave }


{ --------------------------------------------------------- }
{ Jump                                                      }
{ --------------------------------------------------------- }

function TPreprocessor.Jump( over: TFrameSet ) : TFrameMode;
var
  S   : String;
  Key : String;
  Val : String;
  WaterLine: Integer;
begin
  WaterLine := Frame^.Level;
  while Source.GetLine( S ) do
  begin
    SplitPair( S, Key, Val );
    case What(Key) of
      op_endif:
        begin
          Leave;
          if Frame^.Level < WaterLine then
            // on future: 'if fm_endif in over' should be added.
            // now 'over' contains 'fm_endif' anyway.
          begin
            Result := fm_endif;
            Exit;
          end;
        end;

        op_ifdef, op_ifndef:
          Enter( fm_if );

        op_else:
          begin
            if (Frame^.Level = WaterLine) and (fm_else in over) then
            begin
              Leave;
              Result := fm_else;
              Exit;
            end;
          end;
    end;
  end;
  raise Exception.Create( 'Unexpected <eof>' );
end; { Jump }


{ --------------------------------------------------------- }
{ Expand                                                    }
{ --------------------------------------------------------- }

function TPreprocessor.Expand( S: String ) : String;

  procedure Macro( D: PDefine ); far;
  var
    p: Integer;
  begin
    p := Pos( D^.Key, S );
    while p <> 0 do
    begin
      Replace( S, p, Length(D^.Key), D^.Val );
      p := Pos( D^.Key, S );
    end;
  end; { Macro }

begin
  Defines^.ForEach( @Macro );
  Result := S;
end; { Expand }


{ --------------------------------------------------------- }
{ GetLine                                                   }
{ --------------------------------------------------------- }

function TPreprocessor.GetLine( var S: String ) : Boolean;
begin
  repeat
    if Finger < Count then
    begin
      S := Expand( VString(At(Finger))^ );
      Inc( Finger );
      Result := True;
    end
    else
    begin
      S := '';
      Result := False;
    end;
  until not Result or (S <> '');
end; { GetLine }


end.
